(* Renew Zip Archives
   requires command line tool like wzzip.exe (WinZip)

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Jan. 2008

   Command line options:
     /lang:xx         (xx = de, en)
     <filename>       update this zip file
   *)

unit ZipUpdateMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Buttons, HListBox, ComCtrls;

const
  CopRgt = ' 2013 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'info(a)rathlev-home.de';

  NameWinZip = 'WzZip.exe';
  Name7Zip   = '7z.exe';

  PathWinZip = 'WinZip\';
  Path7Zip = '7-Zip\';

type
  TDirInfo = class (TObject)
    DirName     : string;
    SubDir      : boolean;
    constructor Create (FDir : string; ASd : boolean);
    end;

  TfrmZipUpdate = class(TForm)
    pcMain: TPageControl;
    tsRun: TTabSheet;
    tsPrefs: TTabSheet;
    Label1: TLabel;
    btnDir: TSpeedButton;
    btnUpdate: TSpeedButton;
    InfoBtn: TSpeedButton;
    btnExit1: TBitBtn;
    Label2: TLabel;
    btnFile: TSpeedButton;
    OpenDialog: TOpenDialog;
    cbSubDir: TCheckBox;
    edWzZip: TLabeledEdit;
    edUpdate: TLabeledEdit;
    edRecurse: TLabeledEdit;
    btnRun: TBitBtn;
    btnOptions: TBitBtn;
    btnExit2: TBitBtn;
    mmDos: TMemo;
    edFile: TComboBox;
    edDir: TComboBox;
    btnWzip: TBitBtn;
    btn7Zip: TBitBtn;
    edAdd: TLabeledEdit;
    lblAddFile: TLabel;
    Panel1: TPanel;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    btnOther: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure InfoBtnClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btnRunClick(Sender: TObject);
    procedure btnOptionsClick(Sender: TObject);
    procedure btnExit1Click(Sender: TObject);
    procedure btnFileClick(Sender: TObject);
    procedure btnDirClick(Sender: TObject);
    procedure btnUpdateClick(Sender: TObject);
    procedure cbSubDirClick(Sender: TObject);
    procedure edDirCloseUp(Sender: TObject);
    procedure edFileCloseUp(Sender: TObject);
    procedure btnWzipClick(Sender: TObject);
    procedure btn7ZipClick(Sender: TObject);
    procedure btnOtherClick(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    AppPath,UserPath,
    IniName,ProgPath,
    AddFileName         : string;
    AddFile             : boolean;
    procedure UpdateDirList;
    function SetWinZipParams (AskFor : boolean) : boolean;
    function Set7ZipParams (AskFor : boolean) : boolean;
    function GetZipProgram(const APrompt : string; var AName : string) : boolean;
    function SetUserParams : boolean;
  public
    { Public-Deklarationen }
  end;

var
  frmZipUpdate: TfrmZipUpdate;

implementation

{$R *.dfm}

uses JrUtils, ShellDirDlg, IniFiles, GnuGetText, LangUtils, InitProg;

{ ------------------------------------------------------------------- }
constructor TDirInfo.Create (FDir : string; ASd : boolean);
begin
  DirName:=FDir; SubDir:=ASd
  end;

{ ------------------------------------------------------------------- }
const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  ZipSekt = 'ZipFiles';
  DirSekt = 'Directories';

  (* INI-Variablen *)
  iniZip = 'ZipProgram';
  iniSubDir = 'Subdirectories';
  iniUOpt = 'UpdateOption';
  iniROpt = 'RecurseOption';
  iniAOpt = 'AddOption';
  iniFName = 'Name';
  iniFDir = 'Dir';

{ ------------------------------------------------------------------- }
// Combo box list management for directories
const
  Sep = '|';
  mList = 20;

procedure LoadDirListFromIni (IniFile : TIniFile; Section : string;
                              AList: TStrings; MaxCount : integer);
var
  i : integer;
  s : string;
begin
  with IniFile do begin
    if SectionExists(Section) then begin
      AList.Clear;
      for i:=0 to MaxCount-1 do begin
        s:=ReadString(Section,iniFDir+ZStrint(i,2),'');
        if length(s)>0 then AList.Add(s);
        end;
      end;
    end;
  end;

procedure SaveDirListToIni (IniFile : TIniFile; Section : string;
                            Erase : boolean; AList : TStrings; MaxCount : integer);
var
  i,n : integer;
begin
  with IniFile do begin
    if Erase then EraseSection (Section);
    with AList do begin
      if Count>MaxCount then n:=MaxCount else n:=Count;
      for i:=0 to n-1 do
        WriteString(Section,iniFDir+ZStrint(i,2),Strings[i]);
      end;
    end;
  end;

// move or add item "hs" to begin of history list
procedure AddToDirList (AList : TStrings; hs : string; MaxCount : integer);
var
  n : integer;
begin
  if length(hs)>0 then with AList do begin
    n:=IndexOf(hs);
    if n<0 then begin
      if Count>=MaxCount then Delete (Count-1);
      Insert(0,hs);
      end
    else Move (n,0);
    end;
  end;

{ ------------------------------------------------------------------- }
// Combo box list management for zip files
procedure LoadFileListFromIni (IniFile : TIniFile; Section : string;
                              AList: TStrings; MaxCount : integer);
var
  i,n : integer;
  s,t,u : string;
begin
  with IniFile do begin
    if SectionExists(Section) then begin
      AList.Clear;
      for i:=0 to MaxCount-1 do begin
        s:=ReadString(Section,iniFName+ZStrint(i,2),'');
        t:=ReadString(Section,iniFDir+ZStrint(i,2),'');
        if t<>'' then begin
          u:=ReadNxtStr(t,Sep);
          n:=ReadNxtInt(t,Sep,0)
          end
        else begin
          u:=''; n:=0;
          end;
        if length(s)>0 then AList.AddObject(s,TDirInfo.Create(u,boolean(n)));
        end;
      end;
    end;
  end;

procedure SaveFileListToIni (IniFile : TIniFile; Section : string;
                             Erase : boolean; AList : TStrings; MaxCount : integer);
var
  i,n : integer;
begin
  with IniFile do begin
    if Erase then EraseSection (Section);
    with AList do begin
      if Count>MaxCount then n:=MaxCount else n:=Count;
      for i:=0 to n-1 do begin
        WriteString(Section,iniFName+ZStrint(i,2),Strings[i]);
        with (Objects[i] as TDirInfo) do
          WriteString(Section,iniFDir+ZStrint(i,2),DirName+Sep+IntToStr(ord(SubDir)));
        end;
      end;
    end;
  end;

// move or add item "hs" to begin of history list
procedure AddToFileList (AList : TStrings; const hs : string; MaxCount : integer);
var
  n : integer;
begin
  if length(hs)>0 then with AList do begin
    n:=IndexOf(hs);
    if n<0 then begin
      if Count>=MaxCount then Delete (Count-1);
      InsertObject (0,hs,TDirInfo.Create('',false));
      end
    else Move (n,0);
    end;
  end;

procedure UpdateFileList (AList : TStrings; Index : integer; const hd : string; ASd : boolean);
begin
 with (AList.Objects[Index] as TDirInfo) do begin
   DirName:=hd;
   SubDir:=ASd;
   end;
 end;

{ ------------------------------------------------------------------- }
procedure TfrmZipUpdate.FormCreate(Sender: TObject);
var
  IniFile : TIniFile;
  s       : string;
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath,ProgPath);
  InitVersion(_('Refresh Zip Archives'),'Vers. 2.5',CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:=_('Refresh Zip Archives')+' ('+VersInfo.Comments+')';
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    cbSubDir.Checked:=ReadBool(CfgSekt,iniSubDir,false);
    edWzZip.Text:=ReadString(CfgSekt,iniZip,'');
    edUpdate.Text:=ReadString(CfgSekt,iniUOpt,'-f');
    edRecurse.Text:=ReadString(CfgSekt,iniROpt,'-r -p');
    edAdd.Text:=ReadString(CfgSekt,iniAOpt,'');
    LoadFileListFromIni(IniFile,ZipSekt,edFile.Items,mList);
    LoadDirListFromIni(IniFile,DirSekt,edDir.Items,mList);
    Free;
    end;
  edFile.ItemIndex:=-1;
  edDir.ItemIndex:=0;
  AddFile:=false;
  lblAddFile.Caption:='';
  if ParamCount>0 then begin
    with edFile do begin
      s:=ParamStr(1);
      if not AnsiSameText(ExtractFileExt(s),'.zip') then begin
//      if AnsiSameText(s,'add') then begin
        AddFile:=true;
        AddFileName:=s; //ParamStr(2);
        lblAddFile.Caption:=_('Add file: ')+StripPath(AddFileName,80);
        btnUpdate.Caption:=_('Add file');
        end
      else AddToFileList(Items,s,mList);
      ItemIndex:=0;
      end;
    UpdateDirList;
    end;
  cbSubDir.Visible:=not AddFile;  
  end;

procedure TfrmZipUpdate.FormDestroy(Sender: TObject);
var
  IniFile : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    WriteBool(CfgSekt,iniSubDir,cbSubDir.Checked);
    WriteString(CfgSekt,iniZip,edWzZip.Text);
    WriteString(CfgSekt,iniUOpt,edUpdate.Text);
    WriteString(CfgSekt,iniROpt,edRecurse.Text);
    WriteString(CfgSekt,iniAOpt,edAdd.Text);
    SaveFileListToIni(IniFile,ZipSekt,true,edFile.Items,mList);
    SaveDirListToIni(IniFile,DirSekt,true,edDir.Items,mList);
    Free;
    end;
  end;

procedure TfrmZipUpdate.InfoBtnClick(Sender: TObject);
begin
  InfoDialog ('',Caption+' - '+ProgVersDate+sLineBreak+CopRgt
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

procedure TfrmZipUpdate.btnExit1Click(Sender: TObject);
begin
  Close;
  end;

procedure TfrmZipUpdate.FormShow(Sender: TObject);
begin
  if (length(edWzZip.Text)=0) or not FileExists(edWzZip.Text) then begin
    // search for 7-Zip or WinZip
    if not Set7ZipParams(false) and not SetWinZipParams(false) then pcMain.ActivePageIndex:=1
    else pcMain.ActivePageIndex:=0;
    end
  else pcMain.ActivePageIndex:=0;
  end;

procedure TfrmZipUpdate.btnRunClick(Sender: TObject);
begin
  if (length(edWzZip.Text)=0) or not FileExists(edWzZip.Text) then
    ErrorDialog('',_('Zip Command Line Program not found!'))
  else pcMain.ActivePageIndex:=0;
  end;

procedure TfrmZipUpdate.btnOptionsClick(Sender: TObject);
begin
  pcMain.ActivePageIndex:=1;
  end;

function TfrmZipUpdate.GetZipProgram(const APrompt : string; var AName : string) : boolean;
begin
  with OpenDialog do begin
    InitialDir:=ProgPath;
    Filename:=ExtractFilename(AName);
    Filter:=_('Programs|*.exe|all|*.*');
    Title:=Format(_('Select %s command line packer'),[APrompt]);;
    Result:=Execute;
    if Result then AName:=Filename;
    end;
  end;

function TfrmZipUpdate.SetWinZipParams (AskFor : boolean) : boolean;
var
  s : string;
begin
  s:=IncludeTrailingPathDelimiter(ProgPath)+PathWinZip+NameWinZip;
  Result:=FileExists(s);
  if AskFor and not Result then Result:=GetZipProgram('WinZip',s);
  if Result then begin
    edUpdate.Text:='-f';
    edRecurse.Text:='-r -p';
    edAdd.Text:='-u -r -P';
    end;
  end;

function TfrmZipUpdate.Set7ZipParams (AskFor : boolean) : boolean;
var
  s : string;
begin
  s:=IncludeTrailingPathDelimiter(ProgPath)+Path7Zip+Name7Zip;
  Result:=FileExists(s);
  if AskFor and not Result then Result:=GetZipProgram('7-Zip',s);
  if Result then begin
    edWzZip.Text:=s;
    edUpdate.Text:='u -ur0';
    edRecurse.Text:='-r';
    edAdd.Text:='u -r';
    end;
  end;

function TfrmZipUpdate.SetUserParams : boolean;
var
  s : string;
begin
  s:='';
  Result:=GetZipProgram(_('User'),s);
  if Result then begin
    edWzZip.Text:=s;
    edUpdate.Text:='';
    edRecurse.Text:='';
    edAdd.Text:='';
    end;
  end;

procedure TfrmZipUpdate.btnWzipClick(Sender: TObject);
begin
  SetWinZipParams(true);
  end;

procedure TfrmZipUpdate.btn7ZipClick(Sender: TObject);
begin
  Set7ZipParams(true);
  end;

procedure TfrmZipUpdate.btnOtherClick(Sender: TObject);
begin
  SetUserParams;
  end;

procedure TfrmZipUpdate.UpdateDirList;
var
  n : integer;
  s : string;
  b : boolean;
begin
  s:=''; b:=false;
  with edFile do if (Items.Count>0) and (ItemIndex>=0) then
       with (Items.Objects[ItemIndex] as TDirInfo) do begin
    s:=DirName;
    if length(s)>0 then b:=SubDir;
    end;
  if length(s)>0 then with edDir do begin
    n:=Items.IndexOf(s);
    if n>=0 then ItemIndex:=n
    else begin
      AddToDirList(Items,s,mList);
      ItemIndex:=0;
      end;
    end
  else edDir.Text:='';
  cbSubDir.Checked:=b;
  end;

procedure TfrmZipUpdate.btnFileClick(Sender: TObject);
begin
  with OpenDialog do begin
    if length(edFile.Text)>0 then ExtractFilePath(edFile.Text)
    else InitialDir:=UserPath;
    Filename:='';
    Filter:=_('Zip files|*.zip|all|*.*');
    Title:=_('Select Zip file to update');
    if Execute then with edFile do begin
//      Text:=Filename;
      AddToFileList(Items,Filename,mList);
      ItemIndex:=0;
      UpdateDirList;
      end;
    end;
  end;

procedure TfrmZipUpdate.edFileCloseUp(Sender: TObject);
begin
  UpdateDirList;
  end;

procedure TfrmZipUpdate.btnDirClick(Sender: TObject);
var
  s : string;
begin
  s:=edDir.Text;
  if length(s)=0 then s:=ExtractFilePath(edFile.Text);
  if ShellDirDialog.Execute(_('Select directory with files for update'),
                true,true,UserPath,s) then begin
    AddToDirList(edDir.Items,s,mList);
    cbSubDir.State:=cbUnchecked;
    edDir.ItemIndex:=0;
    end;
  end;

procedure TfrmZipUpdate.cbSubDirClick(Sender: TObject);
begin
  with edDir do if Items.Count>0 then Items.Objects[ItemIndex]:=pointer(cbSubDir.State);
  end;

procedure TfrmZipUpdate.edDirCloseUp(Sender: TObject);
begin
  with edDir do if (Items.Count>0) and (ItemIndex>=0) then
    cbSubDir.State:=TCheckBoxState(Items.Objects[ItemIndex]);
  end;

procedure TfrmZipUpdate.btnUpdateClick(Sender: TObject);
const
  BUFSIZE = 1024;
var
  si        : TStartupInfo;
  pi        : TProcessInformation;
  s         : string;
  sa        : AnsiString;
  i         : integer;
  saAttr    : TSecurityAttributes;
  hChildStdoutRd,hChildStdoutWr,
  hChildStdoutRdDup  : THandle;
  chBuf       : array [0..BUFSIZE] of AnsiChar;
  dwRead,ec   : DWord;
begin
  if not FileExists(edFile.Text) then begin
    ErrorDialog('',Format(_('Zip-File not found:'+sLineBreak+'%s'),[edFile.Text]));
    Exit;
    end;
  if not DirectoryExists(edDir.Text) then begin
    ErrorDialog('',Format(_('Directory not found:'+sLineBreak+'%s'),[edDir.Text]));
    Exit;
    end;
  with edFile do UpdateFileList(Items,ItemIndex,edDir.Text,cbSubDir.Checked);
  mmDos.Lines.Clear;
  Application.Processmessages;
  s:=AnsiQuotedStr(edWzZip.Text,'"')+' ';
  if AddFile then begin
    SetCurrentDir(edDir.Text);
    s:=s+edAdd.Text+' '+AnsiQuotedStr(edFile.Text,'"')+' '
                      +AnsiQuotedStr(ExtractRelativePath(IncludeTrailingPathDelimiter(edDir.Text),AddFileName),'"')
    end
  else begin
    s:=s+edUpdate.Text;
    if cbSubDir.Checked then s:=s+' '+edRecurse.Text;
    s:=s+' '+AnsiQuotedStr(edFile.Text,'"')+' '
        +AnsiQuotedStr(IncludeTrailingPathDelimiter(edDir.Text)+'*.*','"');
    end;

// Set the bInheritHandle flag so pipe handles are inherited.
  saAttr.nLength:=sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle:=TRUE;
  saAttr.lpSecurityDescriptor:=nil;

 // The steps for redirecting child process's STDOUT:
 //     1. Save current STDOUT, to be restored later.
 //     2. Create anonymous pipe to be STDOUT for child process.
 //     3. Set STDOUT of the parent process to be write handle to
 //        the pipe, so it is inherited by the child process.
 //        (this is made instead by CreateProcess - StartupInfo, JR)
 //     4. Create a noninheritable duplicate of the read handle and
 //        close the inheritable read handle.

// Create a pipe for the child process's STDOUT.
  if not CreatePipe(hChildStdoutRd,hChildStdoutWr,@saAttr, 0) then begin
    ErrorDialog('',_('Error executing Zip tool')); exit;
    end;
// Create noninheritable read handle and close the inheritable read
// handle.
  if not DuplicateHandle(GetCurrentProcess,hChildStdoutRd,GetCurrentProcess,
                         @hChildStdoutRdDup , 0, FALSE,
                         DUPLICATE_SAME_ACCESS) then begin
    ErrorDialog(Caption,_('Error executing Zip tool')); exit;
    end;
  CloseHandle(hChildStdoutRd);

// Create process to start program
  FillChar(si, SizeOf(TStartupInfo), 0);
  with si do begin
    cb := Sizeof(TStartupInfo);
    dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW;
    wShowWindow:=SW_HIDE;
    hStdOutput :=hChildStdoutWr;
    end;

  Screen.Cursor:=crHourglass;
  if CreateProcess(nil,                // Anwendungsname
                   pchar(s),
                   nil,                // Security
                   nil,                // Security
                   true,               // use InheritHandles
                   NORMAL_PRIORITY_CLASS, // Prioritt
                   nil,                   // Environment
                   nil,                   // Verzeichnis
                   si,pi) then begin
    WaitForSingleObject(pi.hProcess,2000);
    GetExitCodeProcess(pi.hProcess,ec);
    CloseHandle(pi.hProcess);
// Close the write end of the pipe before reading from the
// read end of the pipe.
    if not CloseHandle(hChildStdoutWr) then begin
      ErrorDialog('',_('Error executing Zip tool'));
      Screen.Cursor:=crDefault;
      exit
      end;
    sa:='';
// Read output from the child process, and write to parent's STDOUT.
    while ReadFile(hChildStdoutRdDup,chBuf[0],BUFSIZE,dwRead,nil)
          and (dwRead=BUFSIZE) do begin
      sa:=sa+chBuf;
      end;
    if dwRead>0 then begin
      chBuf[dwread]:=#0;
      sa:=sa+chBuf;
      end;
    CloseHandle(hChildStdoutRdDup);
    // DOS-Ausgabe anzeigen
    s:=sa;
    with mmDos do begin
      mmDos.SetSelTextBuf(pchar(s));
      SelLength:=0;
      Perform(WM_VSCROLL,SB_BOTTOM,0);
      end;
    end
  else begin
    i:=GetLastError;
    ErrorDialog('',Format(_('Error "%u" executing Zip tool'),[i]));
    end;
  Screen.Cursor:=crDefault;
  end;

end.
